package PureMessage::Manager::Users;
use strict;
use warnings;
use Carp;
use PureMessage::Manager::Config qw(%web_config);
use PureMessage::Manager::PasswordGenerator qw(mk_password);
use ActiveState::File::Atomic;
use Exporter;

my @user_functions = qw(
	valid_uid
	valid_password
	add_user
	list_users
	del_user
	change_password
	reset_password
	change_options
);

my @group_functions = qw(
	add_group
	list_groups
	list_user_membership
	list_group_members
	change_groups
	leave_group
	del_group
);

my @acl_functions = qw(
	get_user_acl
	get_merged_user_acl
	can_user_access
	can_group_access
	get_group_acl
	grant_user_acl
	grant_group_acl
	user_has_acl
	sys_get_available_packages	
);

my @common_functions = qw(
	open_db
	commit_db
);

my @all = (@user_functions, @group_functions, @acl_functions, @common_functions);

our $users_err_val;
our $users_err_str;
our %users_err_strs;

our @ISA = qw(Exporter);
our @EXPORT_OK = (@all, '$users_err_str', '$users_err_val', '%users_err_strs');
our %EXPORT_TAGS = ('all'=>   \@all,
		    'user'=>  [@user_functions,   @common_functions],
		    'group'=> [@group_functions,  @common_functions],
		    'acl'=>   [@acl_functions,    @common_functions],
		    );

# The following package variables will eventually move to a
# storage-specific driver.

# use a 3 second timeout when locking files.
my $DB_TIMEOUT = 3;

my %users; # lists users ID's and authentication information
my %groups; # lists groups and each group's ACL
my %membership; # lists user ID's and each group the user belongs to
my %user_acl; # lists user ID's and each user's ACL
# database file names
my $user_db; 
my $group_db;
my $membership_db;
my $user_acl_db;

my $package_dir;
my $opened_ok = 0;

sub valid_password{
	$_ = shift;
	my $uid = shift;
	my $strict = shift;
	if($strict){
		return 0 unless length($_) > 3;
		return 0 unless /[[:punct:]]+/;
		return 0 unless /[\d]+/;
	        return 0 unless ($_ ne $uid);
	}
	return 0 if length($_) > 40;
	return 1;
}

sub valid_uid{
	my $uid = shift;
	croak("You must pass a user ID.") unless defined $uid;
	return 0 if length($uid) > 20;
	return 0 if length($uid) < 1;
	return ($uid =~ /[^a-zA-Z0-9_\-\!\.]+/) ? 0:1;
}

sub list_users{
	return %users;
}

sub list_groups{
	return %groups;
}

sub list_user_membership{
	croak("You must specify a user ID") unless @_;
	unless($users{$_[0]}){
		_set_users_err('uid_missing');
		return 0;
	}
	
	my %ms;
	if($membership{$_[0]}->{'groups'}){
		if(ref($membership{$_[0]}->{'groups'}) eq 'ARRAY'){
			%ms = map { $_=>$_ } @{ $membership{$_[0]}->{'groups'} };
		}else{
			$ms{ $membership{$_[0]}->{'groups'} } = $membership{$_[0]}->{'groups'};
		}
	}
	return %ms;
}

sub list_group_members{
	croak("You must specify a group") unless @_;
	my $group = shift;
	unless($groups{$group}){
		_set_users_err('gid_missing');
		return 0;
	}
	my %members;
	while(my($uid, $minfo) = each(%membership)){
		my $grps = _mkarref($minfo->{'groups'});
		foreach(@$grps){
			next unless $_;
			$members{$uid} = $_ if /$group/;
		}
	}
	return %members;
}

sub add_user{
	my $uid = shift;
	my $password = shift;
	my %options = @_ if @_; # ignored for now
	croak "You must specify a user name and a password"
		unless($uid && defined $password);

	if($users{$uid}){
		_set_users_err('duplicate_uid');
		return 0;
	}

	unless(_valid_uid($uid)){
		_set_users_err('bad_uid');
		return 0;
	}

	unless(valid_password($password, $uid)){
		_set_users_err('bad_pw');
		return 0;
	}
        my $pwd;
        if ($options{crypted}) {
            $pwd = $password;
        }
        else {
            $pwd = $password ? mk_password($password) : "";
        }
        $users{$uid} = {'uid'=>$uid, 'password'=>$pwd, 
                        'name'=>$options{name} || "" ,'passwordv2'=>'yes'};

	return 1;
}

sub add_group{
	my $group = shift;
	my $options = shift if @_;
	croak "You must specify a group"
		unless $group;

	if($groups{$group}){
		_set_users_err('duplicate_gid');
		return 0;
	}
	$groups{$group} = {'group'=>$group};
	$groups{$group}->{'acl'} = $$options{'acl'} if $$options{'acl'};
	return 1;
}

sub change_groups{
	my $uid = shift;
	my $grps = shift;
	croak "You must specify a user ID and a list of groups when changing a group"
		unless($uid && (ref($grps) eq "ARRAY"));

	unless($users{$uid}){
		_set_users_err('uid_missing');
		return 0;
	}
	$membership{$uid}->{'groups'} = $grps;
	$membership{$uid}->{'uid'} = $uid unless $membership{$uid}->{'uid'};
	return 1;
}
sub change_password_crypt{
 	my $uid = shift;
 	my $newp = shift;
 	croak("User ID and new password are all required parameters")
 		unless($uid && $newp);
 	# the new password must be a good one
 	unless(valid_password($newp, $uid)){
 		_set_users_err('bad_pw');
 		return 0;
 	}
 
 	# set the user's new password
 	$users{$uid}->{'password'} = mk_password($newp);
 	$users{$uid}->{'passwordv2'} = 'passwordv2';
 	return 1;
}
sub change_password{
	my $uid = shift;
	my $oldp = shift;
	my $newp = shift;
	croak("User ID, old password and new password are all required parameters")
		unless($uid && $oldp && $newp);

	# the old passwords must match
	my $old_tmp = $users{$uid}->{'password'};
	unless($old_tmp){
		_set_users_err('uid_missing');
		return 0;
	}
	my $passwordv2='no';
    if($users{$uid}->{'passwordv2'}){
        $passwordv2='yes';
    }
 	unless($old_tmp eq mk_password($oldp, $old_tmp,$passwordv2)){
		_set_users_err('old_pw_doesnt_match');
		return 0;
	}

	# the new password must be a good one
	unless(valid_password($newp, $uid)){
		_set_users_err('bad_pw');
		return 0;
	}

	# set the user's new password
	$users{$uid}->{'password'} = mk_password($newp);
	$users{$uid}->{'passwordv2'} = 'passwordv2';
	return 1;
}

sub reset_password{
	my $uid = shift;
	my $newp = shift;

	croak("User ID and new password are both required parameters")
		unless($uid && $newp);

	# the new password must be a good one
	unless(valid_password($newp, $uid)){
		_set_users_err('bad_pw');
		return 0;
	}

	# set the user's new password
	$users{$uid}->{'password'} = mk_password($newp);
	# set the user's new password flag
	$users{$uid}->{'passwordv2'} = 'passwordv2';
	return 1;
}

sub leave_group{
	my $uid = shift;
	my $group = shift;
	croak("User ID and group are required parameters")
		unless($uid && $group);
	unless($groups{$group}){
		_set_users_err('gid_missing');
		return 0;
	}
	unless($users{$uid}){
		_set_users_err('uid_missing');
		return 0;
	}
	my $grps = $membership{$uid}->{'groups'};
	if(ref($grps) eq 'ARRAY'){
		for(my $i = 0; $i < @$grps; $i++){
			splice @$grps, $i, 1 if($$grps[$i] eq $group);
		}
	}else{
		delete $membership{$uid}->{'groups'} if($grps eq $group);
	}
	return 1;
}

sub open_db{
	my $options = shift;
	$user_db = $$options{'user_db'} || $web_config{'user_db'};
	$group_db = $$options{'group_db'} || $web_config{'group_db'};
	$membership_db = $$options{'membership_db'} || $web_config{'membership_db'};
	$user_acl_db = $$options{'user_acl_db'} || $web_config{'user_acl_db'};
	$package_dir = $$options{'package_dir'} || $web_config{'root'};
	
	croak("The user or group databases were not specified.")
		unless($user_db && $group_db);

	# make sure old info is cleared out
	%users = ();
	%groups = ();
	%membership = ();
	%user_acl = ();
	
	# convert any old files if needed
	_convert_old_groups();
	_convert_old_acls();
	
	_read_user_db();
	_read_group_db();
	_read_membership_db();
	_read_user_acl_db();
	$opened_ok = 1;
}

sub commit_db{
	croak("Can't commit if user and group databases are not opened")
		unless $opened_ok;
	return _write_dbs();
}

sub del_user{
	my $uid = shift;
	croak("User ID is a required parameter") unless($uid);
	unless($users{$uid}){
		_set_users_err('uid_missing');
		return 0;
	}
	delete $users{$uid};
	delete $membership{$uid} if $membership{$uid};
	return 1;
}

# this is where it would be nice to have a real database to do this
sub del_group{
	my $group = shift;
	croak("Group is a required parameter") unless($group);
	unless($groups{$group}){
		_set_users_err('gid_missing');
		return 0;
	}
	delete $groups{$group};
	
	# loop through each user's membership and delete all references to the group.
	while ( my ($m,$hr) = each(%membership) ){
		next unless $hr->{'groups'};
		if(ref($$hr{'groups'}) eq 'ARRAY'){
			my $grps = $$hr{'groups'};
			for(my $i = 0; $i < @$grps; $i++){
				delete $$grps[$i] if $$grps[$i] eq $group;
			}
		}else{
			$$hr{'groups'} = 0 if($$hr{'groups'} eq $group);
		}
	}
	return 1;
}

sub get_user_options{
	my $uid = shift;
	croak("User ID is a required parameter") unless($uid);
	return %{ $users{$uid} } if $users{$uid};
	_set_users_err('uid_missing');
	return 0;
}

sub change_user_options{
	my $uid = shift;
	my $opts = shift;
	croak("User ID and user options are required parameters") 
		unless($uid && $opts);

	if($users{$uid}){
		# uid, groups and password are not changeable
		my $p = $users{$uid}->{'password'};
		my $g = $users{$uid}->{'groups'};
		$users{$uid} = $opts;
		$users{$uid}->{'password'} = $p;
		$users{$uid}->{'uid'} = $uid;
		$users{$uid}->{'groups'} = $g;
		return 1;
	}
	_set_users_err('uid_missing');
	return 0;
}

sub user_has_acl{
	my $uid = shift;
	
	# return 0 in call cases. No need to set the error variables
	return 0 unless($users{$uid});
	return 0 unless($user_acl{$uid});
	return 1;
}
sub get_user_acl{
	my $uid = shift;
	croak("User ID is a required parameter") unless $uid;
	unless($user_acl{$uid}){
		# caller should check user_has_acl() first
		_set_users_err('uid_missing_acl');
		return 0;
	}

	# This version only has package names as an array. Future
	# versions of this will likely include more information
	# with each package name.
	my %r;
	foreach ( @{ _mkarref($user_acl{$uid}->{'acl'}) } ) {
	    $r{$_} = $_ if $_;
	}
	return %r;
}

sub get_merged_user_acl{
	my $uid = shift;
	croak("User ID is a required parameter") unless $uid;
	unless($users{$uid}){
		_set_users_err('uid_missing');
		return 0;
	}
	
	# combine the user's ACL with the ACL of any group they belong to
	my @acl = ();
	if($user_acl{$uid} && $user_acl{$uid}->{'acl'}){
		push(@acl, @{ _mkarref($user_acl{$uid}->{'acl'}) });
	}
	
	# get the ACL for each group this user belongs to
	my %gs = list_user_membership($uid);
	foreach my $g(keys %gs){
		croak("Sanity check failed: membership contains a group $g that can't be found.")
			unless($groups{$g});
		next unless $groups{$g}->{'acl'};
		push(@acl, @{ _mkarref($groups{$g}->{'acl'}) });
	}
	
	# Future versions of this will likely include more information
	# with each package name.
	my %r = ();
	%r = map { $_=>$_ } @acl if(@acl);
	return %r;
}
	
sub get_group_acl{
	my $group = shift;
	croak("Group is a required parameter") unless $group;
	unless($groups{$group}){
		_set_users_err('gid_missing');
		return 0;
	}
	
	# Future versions of this will likely include more information
	# with each package name.
	my %r = ();
	%r = map { $_=>$_ } @{ _mkarref($groups{$group}->{'acl'}) }
		if($groups{$group}->{'acl'});
	return %r;
}

sub grant_user_acl{
	my $uid = shift;
	my @acl = @{ $_[0] };
	croak("User ID is a required parameter") unless $uid;
	unless($users{$uid}){
		_set_users_err('uid_missing');
		return 0;
	}

	# clear the user's ACL if an empty list was passed
	unless(@acl){
		$users{$uid}->{'acl'} = 0;
		$user_acl{$uid}->{'acl'} = 0;
		$user_acl{$uid}->{'uid'} = $uid unless $user_acl{$uid}->{'uid'};
		return 1;
	}
	
	# an acl value of '*' give the user access to every package
	if($acl[0] eq '*'){
		$user_acl{$uid}->{'acl'} = '*';
	}else{
		$user_acl{$uid}->{'acl'} = \@acl;
	}
	$user_acl{$uid}->{'uid'} = $uid unless $user_acl{$uid}->{'uid'};
	return 1;
}


sub grant_group_acl{
	my $group = shift;
	my @acl = @{ $_[0] };
	croak("Group is a required parameter") unless $group;
	unless($groups{$group}){
		_set_users_err('gid_missing');
		return 0;
	}

	# an empty list will clear the groups's ACL
	unless(@acl){
		$groups{$group}->{'acl'} = 0;
		return 1;
	}
	
	# an acl value of '*' give the group access to every package
	if($acl[0] eq '*'){
		$groups{$group}->{'acl'} = '*';
	}else{
		$groups{$group}->{'acl'} = \@acl;
	}
	return 1;
}

sub can_user_access{
	my $uid = shift || $ENV{REMOTE_USER};
	my $pkg = shift;
	my %options = @_ if @_;
	croak("User ID and package name are required parameters")
		unless($uid && $pkg);

	# invalid users and packages fail the test
	unless($users{$uid}){
		_set_users_err('uid_missing');
		return 0;
	}
	#my $pkgs = sys_get_available_packages();
	#return 0 unless $$pkgs{$pkg};
	
	# check the user's merged ACL
	# the star acl grants access to everything
	my %acl = get_merged_user_acl($uid);
	return 1 if $acl{'*'};
	return 1 if $acl{$pkg};
	_set_users_err('no_access');
	return 0;
}

sub can_group_access{
	my $group = shift;
	my $pkg = shift;
	my %options = @_ if @_;
	croak("Group and package name are required parameters")
		unless($group && $pkg);
	
	# invalid groups and packages fail the test
	# the star acl grants access to everything
	unless($groups{$group}){
		_set_users_err('gid_missing');
		return 0;
	}
	my %acl = get_group_acl($group);
	return 1 if $acl{'*'};
	return 1 if $acl{$pkg};
	_set_users_err('no_access');
	return 0;
}

# this kind of info could be added to PureMessage::Manager::Config
sub sys_get_available_packages{
	die("The package directory must be set") unless $package_dir;
	die("The package directory cannot be found") unless -d $package_dir;
	
	# Each directory that starts with 'pmx-' is considered a package.
	opendir DH, $package_dir or die "Failed to open the package directory $package_dir. $!";
	my %packages = ();
	foreach my $pkg (readdir(DH)){
		next unless $pkg =~ /^pmx-/;
		$packages{$pkg} = _get_pkg_info("$package_dir/$pkg") if(-d "$package_dir/$pkg");
	}
	closedir DH;
	
	# Pickup the packages listed in the file pmx-okmodules too
	my $pmxok = "$package_dir/pmx-okmodules";
	if(-e $pmxok){
		open FH, "<$pmxok" or die "Failed to open the modules file $pmxok. $!";
		while(<FH>){
			chomp;
			my $pkg = $_;
			if(-d "$package_dir/$pkg"){
				$packages{$pkg} = _get_pkg_info("$package_dir/$pkg");
			}
		}
		close FH;
	}
	return \%packages;
}

# The package's display name, category and sort order are in a file called module.info.
sub _get_pkg_info{
	my $pth = shift;
	croak("Package path is a required parameter") unless $pth;
	
	return 0 unless -e "$pth/module.info";
	open INFO, "<$pth/module.info" or die "Failed to open package info file $pth/module.info. $!";
	my %info = ();
	while(<INFO>){
		chomp;
		next unless /^desc|^category|^order_id/;
		my @line = split '=';
		$info{$line[0]} = $line[1];
	}
	close INFO;
	return \%info;
}

# eventually the following functions will be moved to a local_db driver or something
sub _read_user_db{
	_mk_empty_db($user_db) unless( -e $user_db );
	open FH, "<$user_db" or croak "Failed to open the user database. $!";
	# each record is expected to follow the following format:
	# uid:passwd:[cert]:[allow hosts]|[deny hosts]:name
	# all the optional fields are ignored in this version
	while(<FH>){
		$_ = _trim($_);
		next unless $_;
		my @f = split ':';
		# croak or skip if this record didn't parse? skip for now.
		next unless @f > 1;
		my $passwordv2='no';
        if($users{$f[0]}->{'passwordv2'}){
            $passwordv2='yes';
        }
		my $locked_pwd = "*LK*"; # we need a better API for enabling/disabling
		$users{$f[0]} = {
			'uid' => $f[0], 'password' => $f[1], 
			'enabled' => ($f[1] ne mk_password($locked_pwd, $f[1],$passwordv2)) ? 1 : 0,
                        'name' => $f[4],
			};
		$users{$f[0]}{'passwordv2'} ='passwordv2' if $f[5] && $f[5] eq 'passwordv2';
	}
}

# This leaves temp files lying around if something dies.
#
# FIXME
# There is a race condition in the way this package manages it's data.
# If another process updates the user db between when this process
# read the data and commits the data, the other process's updates
# are lost.
sub _write_dbs{

	my ($uf, $wuf, $gf, $wgf, $mf, $wmf, $af, $waf);
	
	eval{
		$uf = ActiveState::File::Atomic->new($user_db, 'writable'=>1, 'timeout'=>$DB_TIMEOUT);
		$gf = ActiveState::File::Atomic->new($group_db, 'writable'=>1, 'timeout'=>$DB_TIMEOUT);
		$mf = ActiveState::File::Atomic->new($membership_db, 'writable'=>1, 'timeout'=>$DB_TIMEOUT);
		$af = ActiveState::File::Atomic->new($user_acl_db, 'writable'=>1, 'timeout'=>$DB_TIMEOUT);
	};
	if($@){
		croak "Failed to create temporary files. $@";
	}
	
	# save the user, groups, membership and user acl databases
	my $udata = '';
	foreach my $u(keys %users){
		$udata .= "$users{$u}->{'uid'}:";
                $udata .= "$users{$u}->{'password'}:::";
                $udata .= "$users{$u}->{'name'}" if $users{$u}->{'name'};
				$udata .= ":passwordv2" if $users{$u}->{'passwordv2'};
                $udata .= "\n";

	}
		
	my $gdata = '';
	foreach my $g(keys %groups){
		$gdata .= "$groups{$g}->{'group'}:";
		if($groups{$g}->{'acl'}){
			if(ref($groups{$g}->{'acl'}) eq 'ARRAY'){
				$gdata .=  join(' ', @{$groups{$g}->{'acl'}});
			}else{
				$gdata .=  $groups{$g}->{'acl'};
			}
		}
		$gdata .=  "\n";
	}

	my $mdata = '';
	foreach my $m(keys %membership){
		$mdata .= "$membership{$m}->{'uid'}:";
		if($membership{$m}->{'groups'}){
			if(ref($membership{$m}->{'groups'}) eq 'ARRAY'){
				$mdata .= join(' ', @{$membership{$m}->{'groups'}});
			}else{
				$mdata .= $membership{$m}->{'groups'};
			}
		}
		$mdata .= "\n";
	}
	
	my $adata = '';
	foreach my $a(keys %user_acl){
		$adata .= "$user_acl{$a}->{'uid'}:";
		if($user_acl{$a}->{'acl'}){
			if(ref($user_acl{$a}->{'acl'}) eq 'ARRAY'){
				$adata .= join(' ', @{$user_acl{$a}->{'acl'}});
			}else{
				$adata .= $user_acl{$a}->{'acl'};
			}
		}
		$adata .= "\n";
	}
		
	# NOTE: ActiveState::File::Atomic does not support atomic updates of
	# multiple files. If the first file fails, all is ok but if the second
	# or later files fail, all that have been committed will not be reverted.
	#
	# In case of an error, all *open* file handles will be automatically
	# reverted at the end of this function
	$uf->commit_string($udata);
	$gf->commit_string($gdata);
	$mf->commit_string($mdata);
	$af->commit_string($adata);
	return 1;
}

sub _read_group_db{
	
	my @lines = _read_db($group_db);
	foreach my $line(@lines){
		$groups{$$line[0]} =
				{'group'=>$$line[0],
				'acl'=>$$line[1],
				};
	}
}

sub _read_membership_db{
	my @lines = _read_db($membership_db);
	foreach my $line(@lines){
		$membership{$$line[0]} =
				{'uid'=>$$line[0],
				'groups'=>$$line[1],
				}
	}
}

sub _read_user_acl_db{
	# This version only stores package names as an array for a user's
	# acl. Future versions may change this to be a hash for each
	# pacakge so more information can be included.
	my @lines = _read_db($user_acl_db);
	foreach my $line(@lines){
		$user_acl{$$line[0]} =
				{'uid'=>$$line[0], 
				'acl'=>$$line[1],
				};
	}
}

sub _read_db{
	my $db = shift;
	_mk_empty_db($db) unless( -e $db );
	open FH, "<$db" or croak "Failed to open the $db database. $!";
	
	# each record is expected to follow the following format:
	# key:[list_item list_item2]:[list2_item list2_item2] ....
	my @lines = ();
	while(<FH>){
		$_ = _trim($_);
		next unless $_;
		my @r = split ':';
		next unless @r;
		my @line = ();
		foreach(@r){
			# list or scalar
			if(/\s+/){
				# lists are split on white space
				my @tmp = split;
				push(@line, \@tmp);
			}else{
				push(@line, $_);
			}
		}
		push(@lines, \@line);
	}
	close FH;
	return @lines;
}

sub _trim{
	$_ = shift;
	return $_ unless $_;
	s/^\s*//;
	s/\s*$//;
	return $_;
}

sub _valid_uid{
	$_ = shift;
	# This is rather arbitrary and could be made configurable (if needed).
	return 0 unless length($_) < 20;
	return 0 if /[^\w\.\-\!]+/;
	return 1;
}

sub _mkarref{
	my $v = shift;
	if(ref($v) eq 'ARRAY'){
		return $v;
	}else{
		return [$v];
	}
}

sub _set_users_err{
	
	$users_err_val = shift;
	$users_err_str = $users_err_strs{$users_err_val};
}
%users_err_strs =
(
'bad_pw'=>              'Password does not meet the password validation requirements.',
'bad_uid'=>             'User ID does not meet the validation requirements.',
'uid_missing'=>         'User ID does not exist.',
'duplicate_uid'=>       'User already exists.',
'gid_missing'=>         'Group does not exist.',
'duplicate_gid'=>       'Group aready exists.',
'old_pw_doesnt_match'=> 'The password entered did not match the existing password.',
'no_access'=>           'Access denied.',
'uid_missing_acl'=>     'The user does not have an ACL.',
);

sub _mk_empty_db{
	my $db = shift;
	my $mode = shift || 0600;
	
	# don't do anything if the file already exists
	return 1 if (-e $db);
	open(FH, ">$db") or die "Failed to create the db file $db. $!";
	close FH;
	
	# not atomic but will override the umask which will ensure
	# we get the permissions that were specified.
	chmod $mode, $db; 
	return 1;
}

sub _convert_old_groups{
	
	# The old format combinded both membership information and
	# the group's ACL into one record. Break this info appart and
	# for each user in the memberhsip list, add to the membership
	# db. Only the group acl and and membership information, are
	# converted, the other fields are not saved.
	#
	# This process may also create a new membership db. This assums
	# that there shouldn't be a membership db in existance. If there
	# is, it will be backed up; the old and new values are not merged.
		
	# Test the file to see if it is in the old format
	return unless -e $group_db;
	open FH, "<$group_db" or
		die "Failed to open the group database for conversion. $!";
	my $line = join('', (<FH>)) || '';
	close FH;
	chomp $line;
	$line = _trim($line);
	return unless $line;
	return unless $line =~ /^.+:.*:/;
	
	# Convert the file and backup the old version. 
	_mk_empty_db($membership_db) unless -e $membership_db;
	my %opts = ('writable'=>1, 'timeout'=>$DB_TIMEOUT, 'backup_ext'=>'.old');
	my $gf = ActiveState::File::Atomic->new($group_db, %opts);
	my $mf = ActiveState::File::Atomic->new($membership_db, %opts);
	if($gf && $mf){
		my $gwrite = $gf->tempfile;
		my $mwrite = $mf->tempfile;
		$line = '';
		while(defined($line = $gf->readline)){
			print { $gwrite } _convert_group_line($line, $mwrite);
		}
		$gf->commit_tempfile;
		$mf->commit_tempfile;
	}
}

sub _convert_group_line{
	my $line = shift;
	my $mf = shift;
	chomp $line;
	
	# parse out the group acl & membership, ignore the rest
	my @fields = split(':', $line);
	my $gid = $fields[0];
	my $new_line = "$gid:";
	$new_line .= $fields[2] if $fields[2];
	$new_line .= "\n";
	
	# save the new membership if any were included
	if($fields[1]){
		my @members = split(' ', $fields[1]);
		foreach my $member(@members){
			print $mf "$member:$gid\n";
		}
	}
	return $new_line;
}

sub _convert_old_acls{
	# the package acl changed to be pmx-webacl, do a search and replace
	return unless -e $user_acl_db;

	# Convert the file and backup the old version. 
	my %opts = ('writable'=>1, 'timeout'=>$DB_TIMEOUT, 'backup_ext'=>'.old');
	my $af = ActiveState::File::Atomic->new($user_acl_db, %opts);
	my $data = _trim($af->slurp);
	return unless $data;
	$data =~ s/^(.+)\bacl\b/$1pmx-webacl/mg if $data =~ /^(.+)\bacl\b/;
	$af->commit_string($data);
}
1;
__END__

=head1 NAME

PureMessage::Manager::Users - An interface to the Manager user accounts

=head1 SYNOPSIS

  use PureMessage::Manager::Users qw(:all);
  open_db();
  open_db({'user_db'=>'/custom/dir/manager.users'});
  add_user('joe', 'joep', {'group'=>'admin'});
  add_user('jane', 'janep', {'group'=>'admin'});,
  add_group('qadmin');
  add_group('upgrades', {'users'=>['joe', 'jane']});
  change_group('joe', 'qadmin');
  unless(change_password('joe', 'joep', 'peoj')){
	print "Failed to change the password. $users_err_str\n";
  }
  commit_userdb();

=head1 DESCRIPION

The Users module implements the interface used to manage the PureMessage Manager's
user and group databases. The databases may be located in local files or accessed
over the network using some user databse such as LDAP or an SQL database.

=head1 USAGE

The users and groups API.

=over

=item open_userdb([\%options])

This loads the user database so that work can be performed on it. An attempt to
do anything before this has been called will result in an exception. This will
croak if the database cannot be opened.

You can override where to get the user and group databases by specifying 
I<'user_db'> and I<'group_db'> in the C<%options> parameter. The current 
implementation only accepts a file names for these options.

=item commit_userdb

Saves all the changes and closes any open files or open connecitons. This action
is made to be as atomic as possible (depends on the underlying database capabilities)
and will attempt to roll back any changes if there are any errors made. This will
croak if it cannot save the data.

=item add_user($uid, $passwd [,\%options])

Adds a new user to the user database. This will test to make sure that the user
name and password meet certain criteria.

The optional parameter C<\%opions> can be used to specify the default group that
the user belongs to. This can be done by setting the key I<'groups'> to the
name of the group the user belongs to.

=item add_group($group_name [,\%options]) 

Adds a group to the group database.

You can specify a default list of users that belong to the group by setting the
option key I<'usrs'> to an array of user names. 

You can specify the group's ACL by setting the option key I<'acl'> to an array
of ACL hashes. If you do this you can also specify where to get the acl db by
setting the I<'acl_db'> option paramter. The current implementation only accepts
a file name for this option.

This returns false if the group name is invalide. The ACL and user ID's are not
validated. If the ACL is specified this will croak if it cannot open the ACL 
database.

=item user_list

Returns a hash of all the users and their options.

=item group_list

Returns a hash of all the groups. Each group contains, the list of
users that belong to the group, the group's ACL and the group's options.

=item change_group($uid, $group)

Changes the group that a user belongs to the group specified by C<$group>.

=item del_user($uid)

Removes the user from the system (and any groups it belonged to).

Returns false if the user ID did not exist.

=item del_group($group)

Removes the group from the system.

Returns false if the group did not exist.

=item change_password($uid, $old_passwd, $new_passwd)

Change a user's password. C<$old_password> must match the password on the system.
The new password is validated to make sure it meets the same criteria that
I<add_user()> uses.

=item change_user_options($uid, \%options)

Change the options for the user.

Returns false if the user does not exist.

=item get_user_options($uid)

Returns the user's option hash or undefined if the user does not exits.

=back

=cut
